#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "config.h"
#include "imapfilter.h"


extern int sockpri, sockaux;
extern char logfile[PATH_MAX];
extern unsigned int options;

static FILE *logfp = NULL;	/* Pointer to logfile. */


/*
 * Print message if not in OPTION_DETAILS_QUIET mode.
 */
void info(const char *info,...)
{
    va_list args;

    if (options & OPTION_DETAILS_QUIET)
	return;

    va_start(args, info);
    vprintf(info, args);
    va_end(args);
}


/*
 * Print message if in OPTION_DETAILS_VERBOSE mode.
 */
void verbose(const char *info,...)
{
    va_list args;

    if (options & OPTION_DETAILS_VERBOSE) {
	va_start(args, info);
	vprintf(info, args);
	va_end(args);
    }
}


/*
 * Print error message and write it into logfile.
 */
void error(const char *errmsg,...)
{
    va_list args;

    va_start(args, errmsg);

    vfprintf(stderr, errmsg, args);

    if ((options & OPTION_ERRORS) && logfp) {
	vfprintf(logfp, errmsg, args);
    }
    va_end(args);
}


/*
 * Print error message and exit program.
 */
void fatal(unsigned int errnum, const char *fatal,...)
{
    va_list args;

    va_start(args, fatal);
    vfprintf(stderr, fatal, args);
    va_end(args);

    close_connection(&sockpri);
    close_connection(&sockaux);
    secmem_clear();
    tty_restore();
    close_logfile();
    lockfile_remove();

    exit(errnum);
}


/*
 * Catch signals that cause program's termination.
 */
void catch_signals(void)
{
    signal(SIGINT, signal_handler);
    signal(SIGQUIT, signal_handler);
    signal(SIGTERM, signal_handler);
}


/*
 * Signal handler for signals that cause termination of program.
 */
void signal_handler(int sig)
{
    fatal(ERROR_SIGNAL, "imapfilter: killed by signal %d\n", sig);
}


/*
 * Open the file for saving of logging information.
 */
int open_logfile(void)
{
    if (!*logfile)
	return 0;		/* Logging not enabled. */

#ifdef DEBUG
    printf("debug: logfile: '%s'\n", logfile);
#endif

    if (create_file(logfile, S_IRUSR | S_IWUSR))
	return ERROR_TRIVIAL;

    logfp = fopen(logfile, "a");

    if (!logfp) {
	error("imapfilter: opening logfile %s: %s\n",
	      logfile, strerror(errno));
	return ERROR_TRIVIAL;
    }
    return 0;
}


/*
 * Close the logfile.
 */
int close_logfile(void)
{
    if (!logfp)
	return 0;
    else
	return fclose(logfp);
}


/*
 * Prepares the log entry to be saved through continues calls, and writes it
 * to logfile.
 */
void log_info(int flag, void *ptr)
{
    static struct {
	char *account;
	char *mbox;
	char *filter;
	unsigned int *action;
	char *destaccount;
	char *destmbox;
	char *hdrs;
    } inf = {
	NULL, NULL, NULL, NULL, NULL, NULL, NULL
    };

    if (!logfp)
	return;

    switch (flag) {
    case LOG_PREAMBLE:
	fprintf(logfp, "%s %s %s %s %s%s %s\n", get_time(),
		inf.account, inf.mbox, inf.filter,
		(*inf.action == FILTER_ACTION_DELETE ? "delete" :
		 *inf.action == FILTER_ACTION_COPY ? "copy" :
		 *inf.action == FILTER_ACTION_MOVE ? "move" :
		 *inf.action == FILTER_ACTION_RCOPY ? "rcopy " :
		 *inf.action == FILTER_ACTION_RMOVE ? "rmove " :
		 *inf.action == FILTER_ACTION_FLAG_ADD ||
		 *inf.action == FILTER_ACTION_FLAG_REMOVE ||
		 *inf.action == FILTER_ACTION_FLAG_REPLACE ? "flag" :
		 *inf.action == FILTER_ACTION_LIST ? "list" :
		 "unknown"),
		(inf.destaccount == NULL ? "" : inf.destaccount),
		(!*inf.destmbox ? "" : inf.destmbox));
	fflush(logfp);
	break;
    case LOG_ACCOUNT:
	inf.account = (char *)ptr;
	break;
    case LOG_MAILBOX:
	inf.mbox = (char *)ptr;
	break;
    case LOG_FILTER:
	inf.filter = (char *)ptr;
	break;
    case LOG_ACTION:
	inf.action = (unsigned int *)ptr;
	break;
    case LOG_DESTINATION_ACCOUNT:
	if (ptr == NULL)
	    inf.destaccount = NULL;
	else
	    inf.destaccount = ((account_t *) ptr)->key;
	break;
    case LOG_DESTINATION_MAILBOX:
	inf.destmbox = (char *)ptr;
	break;
    case LOG_HEADER:
	    if (ptr) {
		inf.hdrs = (char *)ptr;
		fprintf(logfp, "%c", '\t');
		while (*inf.hdrs) {
		    if (*inf.hdrs == '\r') {
			fprintf(logfp, "%c", '\n');
			if (*(inf.hdrs + 2))
			    fprintf(logfp, "%c", '\t');
			inf.hdrs += 2;
		    } else
			fprintf(logfp, "%c", *(inf.hdrs++));
		}
	    }
	fflush(logfp);
	break;
    }
}



/*
 * Return current local time and date.
 */
char *get_time(void)
{
    char *ct;
    time_t t;

    t = time(NULL);

    ct = ctime(&t);
    *(strchr(ct, '\n')) = 0;

    return ct;
}
