#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <regex.h>
#include <setjmp.h>

#include "config.h"
#include "imapfilter.h"


extern int sockpri;
extern unsigned int capabilities;
extern jmp_buf acctloop;

static char *vbuf = NULL;	/* Virtual buffer. */
static size_t vbufs = 0;	/* Virtual buffer size. */


/*
 * Read one packet of data that the server sent.
 */
void receive_response(int *sock, char *buf)
{
    if (socket_read(sock, buf) == ERROR_NETWORK)
	longjmp(acctloop, -1);

#ifdef DEBUG
    printf("\n%s\n", buf);
#endif

}


/*
 * Get server response to client's request.
 */
int server_response(int *sock, unsigned int tag)
{
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (tag && !strcasestr(buf, ultostr(tag, 16)));

    return analyze_response(buf);
}


/*
 * Check if server sent a BYE response (connection is closed immediately).
 */
void bye_response(char *buf)
{
    if (strcasestr(buf, "* BYE"))
	longjmp(acctloop, -1);
}


/*
 * Process the greeting that server sends during connection.
 */
int greeting_response(int *sock)
{
    char buf[RESPONSE_BUF + 1];

    receive_response(sock, buf);
    bye_response(buf);

    if (strcasestr(buf, "* PREAUTH"))
	return RESPONSE_PREAUTH;

    return RESPONSE_OK;
}


/*
 * Process the data that server sent due to IMAP LOGOUT client request.
 */
int logout_response(int *sock, unsigned int tag)
{
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP CAPABILITY client request.
 */
int capability_response(int *sock, unsigned int tag)
{
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if (!strcasestr(vbuf, "IMAP4rev1")) {
	error("imapfilter: server does not support IMAP4rev1 protocol\n");
	return -2;
    }
    if (strcasestr(vbuf, "NAMESPACE"))
	capabilities |= CAPABILITY_NAMESPACE;

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP NAMESPACE client request.
 */
int namespace_response(int *sock, unsigned int tag, namesp_t * nsp)
{
    char buf[RESPONSE_BUF + 1];
    char *c, *d;

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((c = strcasestr(vbuf, "* NAMESPACE"))) {
	c += 12;
	if (strncasecmp(c, "NIL", 3)) {
	    c = strchr(c, '"') + 1;
	    d = strchr(c, '"') + 1;

	    strncat(nsp->prefix, c, d - c - 1);
	    nsp->delim = *(strchr(d, '"') + 1);
	}
    }
#ifdef DEBUG
    printf("debug: namespace: '%s' '%c'\n", nsp->prefix, nsp->delim);
#endif
    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP STATUS client request.
 */
int status_response(int *sock, unsigned int tag, char *mbox)
{
    int r;
    char buf[RESPONSE_BUF + 1];
    unsigned int exist, recent, unseen;
    char *c;

    exist = recent = unseen = 0;

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    r = analyze_response(buf);

    if (r == RESPONSE_NO)
	return -2;

    if ((c = strcasestr(vbuf, "MESSAGES"))) {
	c += 9;
	exist = strtoul(c, NULL, 10);
    }
    if ((c = strcasestr(vbuf, "RECENT"))) {
	c += 7;
	recent = strtoul(c, NULL, 10);
    }
    if ((c = strcasestr(vbuf, "UNSEEN"))) {
	c += 7;
	unseen = strtoul(c, NULL, 10);
    }
    if (!exist) {
	info("No messages in mailbox \"%s\".\n", mbox);
	return -2;
    }
    info("%d message%s, %d recent, %d unseen, in mailbox \"%s\".\n", exist,
	 plural(exist), recent, unseen, mbox);

    return r;
}


/*
 * Process the data that server sent due to IMAP SELECT client request.
 */
int select_response(int *sock, unsigned int tag)
{
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if (strcasestr(vbuf, "[READ-ONLY]"))
	return RESPONSE_READONLY;

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP SEARCH client request.
 */
int search_response(int *sock, unsigned int tag, char **mesgs)
{
    char buf[RESPONSE_BUF + 1];
    char *c, *m;
    unsigned int blen;

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((c = strcasestr(vbuf, "* SEARCH "))) {
	blen = strlen(vbuf);

	m = *mesgs = (char *)xmalloc(blen + 1);

	c += 9;

	while (*c && (isdigit((unsigned char)(*c)) || *c == ' '))
	    *(m++) = *(c++);

	*m = 0;
    }
    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP FETCH client request.
 */
int fetch_response(int *sock, unsigned int tag, int reset, char *fetch)
{
    char buf[RESPONSE_BUF + 1];
    unsigned int i;
    static unsigned int s;
    char *b;

    if (reset) {
	s = 0;
	return 0;
    }
    i = 0;

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (strlen(vbuf) < RESPONSE_BUF &&
	     !strcasestr(buf, ultostr(tag, 16)));

    b = vbuf;

    if (!s) {
	if ((b = strstr(b, "}\r\n"))) {
	    while (b - vbuf > 0)
		if (*--b == '{')
		    break;
	    s = atoi(++b) - 2;
	    b = strchr(b, '}');
	    b += 3;
	} else {
	    return RESPONSE_NULLBODY;	/* Null body. */
	}
    }
    while (*b && s--)
	fetch[i++] = *(b++);

    fetch[i] = 0;

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP FETCH RFC822.SIZE client
 * request.
 */
int fetchsize_response(int *sock, unsigned int *size, unsigned int tag)
{
    char buf[RESPONSE_BUF + 1];
    char *c;

    *size = 0;

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((c = strcasestr(vbuf, "FETCH (RFC822.SIZE "))) {
	c += 19;
	*size = strtoul(c, NULL, 10);
    }
    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP APPEND client request.
 */
int append_response(int *sock, unsigned int tag)
{
    int r = RESPONSE_OK;
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((r = analyze_response(buf)) == RESPONSE_NO &&
	strcasestr(vbuf, "[TRYCREATE]"))
	return RESPONSE_TRYCREATE;

    return r;
}


/*
 * Process the data that server sent due to IMAP COPY client request.
 */
int copy_response(int *sock, unsigned int tag)
{
    int r = RESPONSE_OK;
    char buf[RESPONSE_BUF + 1];

    reset_vbuf();

    do {
	receive_response(sock, buf);
	bye_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((r = analyze_response(buf)) == RESPONSE_NO &&
	strcasestr(vbuf, "[TRYCREATE]"))
	return RESPONSE_TRYCREATE;

    return r;
}


/*
 * Check if response of server to client's request was succesfully
 * delivered or there was some kind of error.
 */
int analyze_response(char *buf)
{
    int r = RESPONSE_OK;
    regex_t creg;
    regmatch_t match[3];
    const char *reg = "[[:xdigit:]]{6,6} ((OK|NO|BAD)[[:print:]]*)\r\n";
    char result[RESULT_BUF];

    result[0] = 0;

    regcomp(&creg, reg, REG_EXTENDED | REG_ICASE);

    if (!regexec(&creg, buf, 3, match, 0)) {
	strncat(result, buf + match[1].rm_so,
		min(match[1].rm_eo - match[1].rm_so, RESULT_BUF - 1));

	if (!strncasecmp(buf + match[2].rm_so, "NO", 2))
	    r = RESPONSE_NO;
	else if (!strncasecmp(buf + match[2].rm_so, "BAD", 3))
	    r = RESPONSE_BAD;

	verbose("Server response: %s\n", result);
    } else
	r = RESPONSE_NONE;

    regfree(&creg);

    return r;
}


/*
 * Initialize virtual buffer.
 */
void init_vbuf(void)
{
    vbuf = xmalloc(VIRTUAL_BUF);
    *vbuf = 0;
    vbufs = VIRTUAL_BUF;
}


/*
 * Reset virtual buffer.
 */
void reset_vbuf(void)
{
    *vbuf = 0;
}


/*
 * Check if virtual buffer is full and make it bigger.
 */
void check_vbuf(size_t n)
{
    if (n + strlen(vbuf) >= vbufs) {
	vbufs += VIRTUAL_BUF;
	vbuf = xrealloc(vbuf, vbufs);
    }
}
